"""Classes for ABI encoding logic.

Implements classes and functions for serializing Python values into binary data
according to ABI type specifications.
"""
import abc
import codecs
import decimal
from functools import (
    cached_property,
    lru_cache,
)
from types import (
    MethodType,
)
from typing import (
    Any,
    Callable,
    ClassVar,
    Final,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Type,
    final,
)

from faster_eth_utils import (
    is_address,
    is_boolean,
    is_bytes,
    is_integer,
    is_list_like,
    is_number,
    is_text,
    to_canonical_address,
)
from typing_extensions import (
    Self,
)

from faster_eth_abi._encoding import (
    encode_elements,
    encode_elements_dynamic,
    encode_fixed,
    encode_signed,
    encode_tuple,
    encode_tuple_all_dynamic,
    encode_tuple_no_dynamic,
    encode_tuple_no_dynamic_funcs,
    int_to_big_endian,
    validate_tuple,
)
from faster_eth_abi.base import (
    BaseCoder,
)
from faster_eth_abi.exceptions import (
    EncodingTypeError,
    IllegalValue,
    ValueOutOfBounds,
)
from faster_eth_abi.from_type_str import (
    parse_tuple_type_str,
    parse_type_str,
)
from faster_eth_abi.utils.numeric import (
    TEN,
    abi_decimal_context,
    ceil32,
    compute_signed_fixed_bounds,
    compute_signed_integer_bounds,
    compute_unsigned_fixed_bounds,
    compute_unsigned_integer_bounds,
)
from faster_eth_abi.utils.padding import (
    zpad_right,
)
from faster_eth_abi.utils.string import (
    abbr,
)


class BaseEncoder(BaseCoder, metaclass=abc.ABCMeta):
    """
    Base class for all encoder classes.  Subclass this if you want to define a
    custom encoder class.  Subclasses must also implement
    :any:`BaseCoder.from_type_str`.
    """

    @abc.abstractmethod
    def encode(self, value: Any) -> bytes:  # pragma: no cover
        """
        Encodes the given value as a sequence of bytes.  Should raise
        :any:`exceptions.EncodingError` if ``value`` cannot be encoded.
        """

    @abc.abstractmethod
    def validate_value(self, value: Any) -> None:  # pragma: no cover
        """
        Checks whether or not the given value can be encoded by this encoder.
        If the given value cannot be encoded, must raise
        :any:`exceptions.EncodingError`.
        """

    @classmethod
    def invalidate_value(
        cls,
        value: Any,
        exc: Type[Exception] = EncodingTypeError,
        msg: Optional[str] = None,
    ) -> NoReturn:
        """
        Throws a standard exception for when a value is not encodable by an
        encoder.
        """
        raise exc(
            f"Value `{abbr(value)}` of type {type(value)} cannot be encoded by "
            f"{cls.__name__}{'' if msg is None else (': ' + msg)}"
        )

    def __call__(self, value: Any) -> bytes:
        return self.encode(value)


class TupleEncoder(BaseEncoder):
    encoders: Tuple[BaseEncoder, ...] = ()

    def __init__(self, encoders: Tuple[BaseEncoder, ...], **kwargs: Any) -> None:
        super().__init__(encoders=encoders, **kwargs)

        self._is_dynamic: Final = tuple(
            getattr(e, "is_dynamic", False) for e in self.encoders
        )
        self.is_dynamic = any(self._is_dynamic)

        validators = []
        for encoder in self.encoders:
            try:
                validator = encoder.validate_value
            except AttributeError:
                validators.append(encoder)
            else:
                validators.append(validator)

        self.validators: Final[Callable[[Any], None]] = tuple(validators)

        if type(self).encode is TupleEncoder.encode:
            encode_func = (
                encode_tuple_all_dynamic
                if all(self._is_dynamic)
                else encode_tuple_no_dynamic_funcs.get(
                    len(self.encoders),
                    encode_tuple_no_dynamic,
                )
                if not self.is_dynamic
                else encode_tuple
            )

            self.encode = MethodType(encode_func, self)

    def validate(self) -> None:
        super().validate()

        if self.encoders is None:
            raise ValueError("`encoders` may not be none")

    @final
    def validate_value(self, value: Sequence[Any]) -> None:
        validate_tuple(self, value)

    def encode(self, values: Sequence[Any]) -> bytes:
        return encode_tuple(self, values)

    def __call__(self, values: Sequence[Any]) -> bytes:
        return self.encode(values)

    @parse_tuple_type_str
    def from_type_str(cls, abi_type, registry):
        encoders = tuple(
            registry.get_encoder(c.to_type_str()) for c in abi_type.components
        )

        return cls(encoders=encoders)


class FixedSizeEncoder(BaseEncoder):
    value_bit_size = None
    data_byte_size = None
    encode_fn = None
    type_check_fn = None
    is_big_endian = None

    def validate(self) -> None:
        super().validate()

        value_bit_size = self.value_bit_size
        if value_bit_size is None:
            raise ValueError("`value_bit_size` may not be none")
        data_byte_size = self.data_byte_size
        if data_byte_size is None:
            raise ValueError("`data_byte_size` may not be none")
        if self.encode_fn is None:
            raise ValueError("`encode_fn` may not be none")
        if self.is_big_endian is None:
            raise ValueError("`is_big_endian` may not be none")

        if value_bit_size % 8 != 0:
            raise ValueError(
                f"Invalid value bit size: {value_bit_size}. Must be a multiple of 8"
            )

        if value_bit_size > data_byte_size * 8:
            raise ValueError("Value byte size exceeds data size")

    def validate_value(self, value):
        raise NotImplementedError("Must be implemented by subclasses")

    def encode(self, value: Any) -> bytes:
        self.validate_value(value)
        encode_fn = self.encode_fn
        if encode_fn is None:
            raise AssertionError("`encode_fn` is None")
        return encode_fixed(value, encode_fn, self.is_big_endian, self.data_byte_size)

    __call__ = encode


class Fixed32ByteSizeEncoder(FixedSizeEncoder):
    data_byte_size = 32


class BooleanEncoder(Fixed32ByteSizeEncoder):
    value_bit_size = 8
    is_big_endian = True

    @classmethod
    def validate_value(cls, value: Any) -> None:
        if not is_boolean(value):
            cls.invalidate_value(value)

    @classmethod
    def encode_fn(cls, value: bool) -> bytes:
        if value is True:
            return b"\x01"
        elif value is False:
            return b"\x00"
        else:
            raise ValueError("Invariant")

    @parse_type_str("bool")
    def from_type_str(cls, abi_type, registry):
        return cls()


class PackedBooleanEncoder(BooleanEncoder):
    data_byte_size = 1


class NumberEncoder(Fixed32ByteSizeEncoder):
    is_big_endian = True
    bounds_fn = None
    illegal_value_fn = None
    type_check_fn = None

    def validate(self) -> None:
        super().validate()

        if self.bounds_fn is None:
            raise ValueError("`bounds_fn` cannot be null")
        if self.type_check_fn is None:
            raise ValueError("`type_check_fn` cannot be null")

    def validate_value(self, value):
        type_check_fn = self.type_check_fn
        if type_check_fn is None:
            raise AssertionError("`type_check_fn` is None")
        if not type_check_fn(value):
            self.invalidate_value(value)

        illegal_value_fn = self.illegal_value_fn
        illegal_value = illegal_value_fn is not None and illegal_value_fn(value)
        if illegal_value:
            self.invalidate_value(value, exc=IllegalValue)

        lower_bound, upper_bound = self.bounds_fn(self.value_bit_size)
        if value < lower_bound or value > upper_bound:
            self.invalidate_value(
                value,
                exc=ValueOutOfBounds,
                msg=f"Cannot be encoded in {self.value_bit_size} bits. Must be bounded "
                f"between [{lower_bound}, {upper_bound}].",
            )


class UnsignedIntegerEncoder(NumberEncoder):
    encode_fn = staticmethod(int_to_big_endian)
    bounds_fn = staticmethod(compute_unsigned_integer_bounds)
    type_check_fn = staticmethod(is_integer)

    @parse_type_str("uint")
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub)


class UnsignedIntegerEncoderCached(UnsignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[Optional[int]]

    def __init__(self, maxsize: Optional[int] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.maxsize = maxsize
        self.encode = lru_cache(maxsize=maxsize)(self.encode)


encode_uint_256 = UnsignedIntegerEncoder(value_bit_size=256, data_byte_size=32)


class PackedUnsignedIntegerEncoder(UnsignedIntegerEncoder):
    @parse_type_str("uint")
    def from_type_str(cls, abi_type, registry):
        return cls(
            value_bit_size=abi_type.sub,
            data_byte_size=abi_type.sub // 8,
        )


class PackedUnsignedIntegerEncoderCached(PackedUnsignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[Optional[int]]

    def __init__(self, maxsize: Optional[int] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.maxsize = maxsize
        self.encode = lru_cache(maxsize=maxsize)(self.encode)


class SignedIntegerEncoder(NumberEncoder):
    bounds_fn = staticmethod(compute_signed_integer_bounds)
    type_check_fn = staticmethod(is_integer)

    def encode_fn(self, value: int) -> bytes:
        return int_to_big_endian(value % (2**self.value_bit_size))

    def encode(self, value: int) -> bytes:
        self.validate_value(value)
        return encode_signed(value, self.encode_fn, self.data_byte_size)

    __call__ = encode

    @parse_type_str("int")
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub)


class SignedIntegerEncoderCached(SignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[Optional[int]]

    def __init__(self, maxsize: Optional[int] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.maxsize = maxsize
        self.encode = lru_cache(maxsize=maxsize)(self.encode)


class PackedSignedIntegerEncoder(SignedIntegerEncoder):
    @parse_type_str("int")
    def from_type_str(cls, abi_type, registry):
        return cls(
            value_bit_size=abi_type.sub,
            data_byte_size=abi_type.sub // 8,
        )


class PackedSignedIntegerEncoderCached(PackedSignedIntegerEncoder):
    encode: Final[Callable[[int], bytes]]
    maxsize: Final[Optional[int]]

    def __init__(self, maxsize: Optional[int] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.maxsize = maxsize
        self.encode = lru_cache(maxsize=maxsize)(self.encode)


class BaseFixedEncoder(NumberEncoder):
    frac_places = None

    @staticmethod
    def type_check_fn(value):
        return is_number(value) and not isinstance(value, float)

    @staticmethod
    def illegal_value_fn(value):
        if isinstance(value, decimal.Decimal):
            return value.is_nan() or value.is_infinite()

        return False

    @cached_property
    def denominator(self) -> decimal.Decimal:
        return TEN**self.frac_places

    @cached_property
    def precision(self) -> int:
        return TEN**-self.frac_places

    def validate_value(self, value):
        super().validate_value(value)

        with decimal.localcontext(abi_decimal_context):
            residue = value % self.precision

        if residue > 0:
            self.invalidate_value(
                value,
                exc=IllegalValue,
                msg=f"residue {residue!r} outside allowed fractional precision of "
                f"{self.frac_places}",
            )

    def validate(self) -> None:
        super().validate()

        frac_places = self.frac_places
        if frac_places is None:
            raise ValueError("must specify `frac_places`")

        if frac_places <= 0 or frac_places > 80:
            raise ValueError("`frac_places` must be in range (0, 80]")


class UnsignedFixedEncoder(BaseFixedEncoder):
    def bounds_fn(self, value_bit_size):
        return compute_unsigned_fixed_bounds(self.value_bit_size, self.frac_places)

    def encode_fn(self, value):
        with decimal.localcontext(abi_decimal_context):
            scaled_value = value * self.denominator
            integer_value = int(scaled_value)

        return int_to_big_endian(integer_value)

    @parse_type_str("ufixed")
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(
            value_bit_size=value_bit_size,
            frac_places=frac_places,
        )


class PackedUnsignedFixedEncoder(UnsignedFixedEncoder):
    @parse_type_str("ufixed")
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(
            value_bit_size=value_bit_size,
            data_byte_size=value_bit_size // 8,
            frac_places=frac_places,
        )


class SignedFixedEncoder(BaseFixedEncoder):
    def bounds_fn(self, value_bit_size):
        return compute_signed_fixed_bounds(self.value_bit_size, self.frac_places)

    def encode_fn(self, value):
        with decimal.localcontext(abi_decimal_context):
            scaled_value = value * self.denominator
            integer_value = int(scaled_value)

        unsigned_integer_value = integer_value % (2**self.value_bit_size)

        return int_to_big_endian(unsigned_integer_value)

    def encode(self, value):
        self.validate_value(value)
        return encode_signed(value, self.encode_fn, self.data_byte_size)

    __call__ = encode

    @parse_type_str("fixed")
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(
            value_bit_size=value_bit_size,
            frac_places=frac_places,
        )


class PackedSignedFixedEncoder(SignedFixedEncoder):
    @parse_type_str("fixed")
    def from_type_str(cls, abi_type, registry):
        value_bit_size, frac_places = abi_type.sub

        return cls(
            value_bit_size=value_bit_size,
            data_byte_size=value_bit_size // 8,
            frac_places=frac_places,
        )


class AddressEncoder(Fixed32ByteSizeEncoder):
    value_bit_size = 20 * 8
    encode_fn = staticmethod(to_canonical_address)
    is_big_endian = True

    @classmethod
    def validate_value(cls, value: Any) -> None:
        if not is_address(value):
            cls.invalidate_value(value)

    def validate(self) -> None:
        super().validate()

        if self.value_bit_size != 20 * 8:
            raise ValueError("Addresses must be 160 bits in length")

    @parse_type_str("address")
    def from_type_str(cls, abi_type, registry):
        return cls()


class PackedAddressEncoder(AddressEncoder):
    data_byte_size = 20


class BytesEncoder(Fixed32ByteSizeEncoder):
    is_big_endian = False

    def validate_value(self, value: Any) -> None:
        if not is_bytes(value):
            self.invalidate_value(value)

        byte_size = self.value_bit_size // 8
        if len(value) > byte_size:
            self.invalidate_value(
                value,
                exc=ValueOutOfBounds,
                msg=f"exceeds total byte size for bytes{byte_size} encoding",
            )

    @staticmethod
    def encode_fn(value):
        return value

    @parse_type_str("bytes")
    def from_type_str(cls, abi_type, registry):
        return cls(value_bit_size=abi_type.sub * 8)


class PackedBytesEncoder(BytesEncoder):
    @parse_type_str("bytes")
    def from_type_str(cls, abi_type, registry):
        return cls(
            value_bit_size=abi_type.sub * 8,
            data_byte_size=abi_type.sub,
        )


class ByteStringEncoder(BaseEncoder):
    is_dynamic = True

    @classmethod
    def validate_value(cls, value: Any) -> None:
        if not is_bytes(value):
            cls.invalidate_value(value)

    @classmethod
    def encode(cls, value: bytes) -> bytes:
        cls.validate_value(value)
        value_length = len(value)

        encoded_size = encode_uint_256(value_length)
        padded_value = zpad_right(value, ceil32(value_length))

        return encoded_size + padded_value

    __call__: ClassVar[Callable[[Type[Self], bytes], bytes]] = encode

    @parse_type_str("bytes")
    def from_type_str(cls, abi_type, registry):
        return cls()  # type: ignore [misc]


class PackedByteStringEncoder(ByteStringEncoder):
    is_dynamic = False

    @classmethod
    def encode(cls, value):
        cls.validate_value(value)
        return value

    __call__ = encode


class TextStringEncoder(BaseEncoder):
    is_dynamic = True

    @classmethod
    def validate_value(cls, value: Any) -> None:
        if not is_text(value):
            cls.invalidate_value(value)

    @classmethod
    def encode(cls, value: str) -> bytes:
        cls.validate_value(value)

        value_as_bytes = codecs.encode(value, "utf8")
        value_length = len(value_as_bytes)

        encoded_size = encode_uint_256(value_length)
        padded_value = zpad_right(value_as_bytes, ceil32(value_length))

        return encoded_size + padded_value

    __call__: ClassVar[Callable[[Type[Self], str], bytes]] = encode

    @parse_type_str("string")
    def from_type_str(cls, abi_type, registry):
        return cls()  # type: ignore [misc]


class PackedTextStringEncoder(TextStringEncoder):
    is_dynamic = False

    @classmethod
    def encode(cls, value: str) -> bytes:
        cls.validate_value(value)
        return codecs.encode(value, "utf8")

    __call__ = encode


class BaseArrayEncoder(BaseEncoder):
    item_encoder: BaseEncoder = None

    def validate(self) -> None:
        super().validate()

        if self.item_encoder is None:
            raise ValueError("`item_encoder` may not be none")

    def validate_value(self, value: Any) -> None:
        if not is_list_like(value):
            self.invalidate_value(
                value,
                msg="must be list-like such as array or tuple",
            )

        item_encoder = self.item_encoder
        for item in value:
            item_encoder.validate_value(item)

    def encode_elements(self, value: Sequence[Any]) -> bytes:
        self.validate_value(value)
        return encode_elements(self.item_encoder, value)

    @parse_type_str(with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        item_encoder = registry.get_encoder(abi_type.item_type.to_type_str())

        array_spec = abi_type.arrlist[-1]
        if len(array_spec) == 1:
            # If array dimension is fixed
            return SizedArrayEncoder(
                array_size=array_spec[0],
                item_encoder=item_encoder,
            )
        else:
            # If array dimension is dynamic
            return DynamicArrayEncoder(item_encoder=item_encoder)


class PackedArrayEncoder(BaseArrayEncoder):
    array_size = None

    def validate_value(self, value: Any) -> None:
        super().validate_value(value)

        array_size = self.array_size
        if array_size is not None and len(value) != array_size:
            self.invalidate_value(
                value,
                exc=ValueOutOfBounds,
                msg=f"value has {len(value)} items when {array_size} were expected",
            )

    def encode(self, value: Sequence[Any]) -> bytes:
        return encode_elements(self.item_encoder, value)

    __call__ = encode

    @parse_type_str(with_arrlist=True)
    def from_type_str(cls, abi_type, registry):
        item_encoder = registry.get_encoder(abi_type.item_type.to_type_str())

        array_spec = abi_type.arrlist[-1]
        if len(array_spec) == 1:
            return cls(
                array_size=array_spec[0],
                item_encoder=item_encoder,
            )
        else:
            return cls(item_encoder=item_encoder)


class SizedArrayEncoder(BaseArrayEncoder):
    array_size = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.is_dynamic = self.item_encoder.is_dynamic

    def validate(self) -> None:
        super().validate()

        if self.array_size is None:
            raise ValueError("`array_size` may not be none")

    def validate_value(self, value: Any) -> None:
        super().validate_value(value)

        if len(value) != self.array_size:
            self.invalidate_value(
                value,
                exc=ValueOutOfBounds,
                msg=f"value has {len(value)} items when {self.array_size} were "
                "expected",
            )

    def encode(self, value: Sequence[Any]) -> bytes:
        return encode_elements(self.item_encoder, value)

    __call__ = encode


class DynamicArrayEncoder(BaseArrayEncoder):
    is_dynamic = True

    def encode(self, value: Sequence[Any]) -> bytes:
        return encode_elements_dynamic(self.item_encoder, value)



"""
compare_benchmark_results.py

Compares the two implementations in each benchmark group from a pytest-benchmark parsed results JSON file,
grouped by submodule.
For each submodule and group (e.g., "abi_to_signature"), finds both implementations
(e.g., "test_abi_to_signature" and "test_faster_abi_to_signature"), computes the percent change
in mean execution time, speedup percent, and x factor, and writes a diff JSON file summarizing the results.

Usage:
    python compare_benchmark_results.py <results.json> [output.json]
"""

import json
import re
import sys
from typing import (
    Any,
    Dict,
)


def get_group_name(test_name: str) -> str:
    # Extract group from test name, e.g., test_foo, test_faster_foo -> group: foo
    m = re.match(r"test_faster_(.+)", test_name)
    if m:
        return m.group(1)
    m = re.match(r"test_(.+)", test_name)
    if m:
        return m.group(1)
    return test_name


def compare_group(group_results: Dict[str, Any]) -> Dict[str, Any]:
    # Find reference and faster implementations in the group
    ref = None
    fast = None
    ref_name = None
    fast_name = None
    for func_name, data in group_results.items():
        if func_name.startswith("test_faster_"):
            fast = data
            fast_name = func_name
        elif func_name.startswith("test_"):
            ref = data
            ref_name = func_name

    if ref and fast:
        mean_ref = ref["mean"]
        mean_fast = fast["mean"]
        percent_change = (
            ((mean_ref - mean_fast) / mean_ref) * 100 if mean_ref != 0 else 0.0
        )
        speedup_x = mean_ref / mean_fast if mean_fast != 0 else float("inf")
        speedup_percent = (
            (speedup_x - 1) * 100 if speedup_x != float("inf") else float("inf")
        )
        return {
            "reference_mean": mean_ref,
            "faster_mean": mean_fast,
            "percent_change": percent_change,
            "speedup_percent": speedup_percent,
            "speedup_x": speedup_x,
            "reference": ref_name,
            "faster": fast_name,
        }
    else:
        missing = []
        if not ref:
            missing.append("reference")
        if not fast:
            missing.append("faster")
        return {"note": f"Missing implementation(s): {missing}"}


def main() -> None:
    if len(sys.argv) < 2:
        print("Usage: python compare_benchmark_results.py <results.json> [output.json]")
        sys.exit(1)
    results_path = sys.argv[1]
    output_path = sys.argv[2] if len(sys.argv) > 2 else "benchmark_diff.json"

    with open(results_path) as f:
        results = json.load(f)

    # results: {submodule: {group: {function_name: {...}}}}
    diff_by_submodule = {
        submodule: {
            group: compare_group(group_results)
            for group, group_results in groups.items()
        }
        for submodule, groups in results.items()
    }

    with open(output_path, "w") as f:
        json.dump(diff_by_submodule, f, indent=2)
    print(f"Diff written to {output_path}")


if __name__ == "__main__":
    main()



"""Helpers for parsing and normalizing ABI type strings.

Provides decorators and utilities for implementing from_type_str methods on coder classes.
"""
import functools
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Type,
    TypeVar,
)

from eth_typing import (
    TypeStr,
)

from ._grammar import (
    ABIType,
    BasicType,
    TupleType,
    normalize,
)
from .grammar import (
    parse,
)

if TYPE_CHECKING:
    from .base import (
        BaseCoder,
    )


TType = TypeVar("TType", bound=Type["BaseCoder"])
OldFromTypeStr = Callable[["BaseCoder", ABIType, Any], TType]
if TYPE_CHECKING:
    NewFromTypeStr = classmethod[TType, [TypeStr, Any], TType]


def parse_type_str(
    expected_base: Optional[str] = None,
    with_arrlist: bool = False,
) -> Callable[[OldFromTypeStr[TType]], "NewFromTypeStr[TType]"]:
    """
    Used by BaseCoder subclasses as a convenience for implementing the
    ``from_type_str`` method required by ``ABIRegistry``.  Useful if normalizing
    then parsing a type string with an (optional) expected base is required in
    that method.
    """

    def decorator(old_from_type_str: OldFromTypeStr[TType]) -> "NewFromTypeStr[TType]":
        @functools.wraps(old_from_type_str)
        def new_from_type_str(cls: TType, type_str: TypeStr, registry: Any) -> TType:
            normalized_type_str = normalize(type_str)
            abi_type = parse(normalized_type_str)

            type_str_repr = repr(type_str)
            if type_str != normalized_type_str:
                type_str_repr = (
                    f"{type_str_repr} (normalized to {normalized_type_str!r})"
                )

            if expected_base is not None:
                if not isinstance(abi_type, BasicType):
                    raise ValueError(
                        "Cannot create {} for non-basic type {}".format(
                            cls.__name__,
                            type_str_repr,
                        )
                    )
                if abi_type.base != expected_base:
                    raise ValueError(
                        "Cannot create {} for type {}: expected type with "
                        "base '{}'".format(
                            cls.__name__,
                            type_str_repr,
                            expected_base,
                        )
                    )

            if not with_arrlist and abi_type.arrlist is not None:
                raise ValueError(
                    "Cannot create {} for type {}: expected type with "
                    "no array dimension list".format(
                        cls.__name__,
                        type_str_repr,
                    )
                )
            if with_arrlist and abi_type.arrlist is None:
                raise ValueError(
                    "Cannot create {} for type {}: expected type with "
                    "array dimension list".format(
                        cls.__name__,
                        type_str_repr,
                    )
                )

            # Perform general validation of default solidity types
            abi_type.validate()

            return old_from_type_str(cls, abi_type, registry)

        return classmethod(new_from_type_str)

    return decorator


def parse_tuple_type_str(
    old_from_type_str: OldFromTypeStr[TType],
) -> "NewFromTypeStr[TType]":
    """
    Used by BaseCoder subclasses as a convenience for implementing the
    ``from_type_str`` method required by ``ABIRegistry``.  Useful if normalizing
    then parsing a tuple type string is required in that method.
    """

    @functools.wraps(old_from_type_str)
    def new_from_type_str(cls: TType, type_str: TypeStr, registry: Any) -> TType:
        normalized_type_str = normalize(type_str)
        abi_type = parse(normalized_type_str)

        if not isinstance(abi_type, TupleType):
            type_str_repr = repr(type_str)
            if type_str != normalized_type_str:
                type_str_repr = "{} (normalized to {})".format(
                    type_str_repr,
                    repr(normalized_type_str),
                )
            raise ValueError(
                "Cannot create {} for non-tuple type {}".format(
                    cls.__name__,
                    type_str_repr,
                )
            )

        abi_type.validate()

        return old_from_type_str(cls, abi_type, registry)

    return classmethod(new_from_type_str)



# AUTOGENERATED: Run scripts/generate_overload_tests.py to update.
# Requires mypy extras
# To run: pip install mypy && mypy tests/typecheck

# alias some things so we can minimize test data size in source control
from typing import Any as A
from typing import Tuple as T
from typing import Union as U
from typing_extensions import assert_type as ck
from eth_typing import HexAddress as HA
from faster_eth_abi import abi

# more aliases to save a bunch more chars
bo = bool
by = bytes
i = int
s = str

TA = T[A, ...]

THA = T[HA, ...]
Tbo = T[bo, ...]
Tby = T[by, ...]
Ti = T[i, ...]
Ts = T[s, ...]
TUa_bo = T[U[HA, bo], ...]
TUa_by = T[U[HA, by], ...]
TUa_i = T[U[HA, i], ...]
TUa_s = T[U[HA, s], ...]
TUbo_by = T[U[bo, by], ...]
TUbo_i = T[U[bo, i], ...]
TUbo_s = T[U[bo, s], ...]
TUby_i = T[U[by, i], ...]
TUby_s = T[U[by, s], ...]
TUi_s = T[U[i, s], ...]
TUa_bo_by = T[U[bo, by, HA], ...]
TUbo_by_i = T[U[bo, by, i], ...]
TUbo_by_s = T[U[bo, by, s], ...]
TUa_bo_i = T[U[bo, i, HA], ...]
TUa_by_i = T[U[by, i, HA], ...]
TUby_i_s = T[U[by, i, s], ...]

# the callable we are checking
f = abi.decode

# input data
D = b"\x00" * 32
ck(f(['address'], D), THA)  # var case 0
ck(f(['address[]'], D), TA)  # var case 1
ck(f(['bool'], D), Tbo)  # var case 2
ck(f(['bool[]'], D), TA)  # var case 3
ck(f(['bytes'], D), Tby)  # var case 4
ck(f(['bytes1'], D), Tby)  # var case 5
ck(f(['bytes10'], D), Tby)  # var case 6
ck(f(['bytes11'], D), Tby)  # var case 7
ck(f(['bytes12'], D), Tby)  # var case 8
ck(f(['bytes13'], D), Tby)  # var case 9
ck(f(['bytes14'], D), Tby)  # var case 10
ck(f(['bytes15'], D), Tby)  # var case 11
ck(f(['bytes16'], D), Tby)  # var case 12
ck(f(['bytes17'], D), Tby)  # var case 13
ck(f(['bytes18'], D), Tby)  # var case 14
ck(f(['bytes19'], D), Tby)  # var case 15
ck(f(['bytes2'], D), Tby)  # var case 16
ck(f(['bytes20'], D), Tby)  # var case 17
ck(f(['bytes21'], D), Tby)  # var case 18
ck(f(['bytes22'], D), Tby)  # var case 19
ck(f(['bytes23'], D), Tby)  # var case 20
ck(f(['bytes24'], D), Tby)  # var case 21
ck(f(['bytes25'], D), Tby)  # var case 22
ck(f(['bytes26'], D), Tby)  # var case 23
ck(f(['bytes27'], D), Tby)  # var case 24
ck(f(['bytes28'], D), Tby)  # var case 25
ck(f(['bytes29'], D), Tby)  # var case 26
ck(f(['bytes3'], D), Tby)  # var case 27
ck(f(['bytes30'], D), Tby)  # var case 28
ck(f(['bytes31'], D), Tby)  # var case 29
ck(f(['bytes32'], D), Tby)  # var case 30
ck(f(['bytes4'], D), Tby)  # var case 31
ck(f(['bytes5'], D), Tby)  # var case 32
ck(f(['bytes6'], D), Tby)  # var case 33
ck(f(['bytes7'], D), Tby)  # var case 34
ck(f(['bytes8'], D), Tby)  # var case 35
ck(f(['bytes9'], D), Tby)  # var case 36
ck(f(['int104'], D), Ti)  # var case 37
ck(f(['int112'], D), Ti)  # var case 38
ck(f(['int120'], D), Ti)  # var case 39
ck(f(['int128'], D), Ti)  # var case 40
ck(f(['int136'], D), Ti)  # var case 41
ck(f(['int144'], D), Ti)  # var case 42
ck(f(['int152'], D), Ti)  # var case 43
ck(f(['int16'], D), Ti)  # var case 44
ck(f(['int160'], D), Ti)  # var case 45
ck(f(['int168'], D), Ti)  # var case 46
ck(f(['int176'], D), Ti)  # var case 47
ck(f(['int184'], D), Ti)  # var case 48
ck(f(['int192'], D), Ti)  # var case 49
ck(f(['int200'], D), Ti)  # var case 50
ck(f(['int208'], D), Ti)  # var case 51
ck(f(['int216'], D), Ti)  # var case 52
ck(f(['int224'], D), Ti)  # var case 53
ck(f(['int232'], D), Ti)  # var case 54
ck(f(['int24'], D), Ti)  # var case 55
ck(f(['int240'], D), Ti)  # var case 56
ck(f(['int248'], D), Ti)  # var case 57
ck(f(['int256'], D), Ti)  # var case 58
ck(f(['int32'], D), Ti)  # var case 59
ck(f(['int40'], D), Ti)  # var case 60
ck(f(['int48'], D), Ti)  # var case 61
ck(f(['int56'], D), Ti)  # var case 62
ck(f(['int64'], D), Ti)  # var case 63
ck(f(['int72'], D), Ti)  # var case 64
ck(f(['int8'], D), Ti)  # var case 65
ck(f(['int80'], D), Ti)  # var case 66
ck(f(['int88'], D), Ti)  # var case 67
ck(f(['int96'], D), Ti)  # var case 68
ck(f(['string'], D), Ts)  # var case 69
ck(f(['string[]'], D), TA)  # var case 70
ck(f(['uint104'], D), Ti)  # var case 71
ck(f(['uint112'], D), Ti)  # var case 72
ck(f(['uint120'], D), Ti)  # var case 73
ck(f(['uint128'], D), Ti)  # var case 74
ck(f(['uint136'], D), Ti)  # var case 75
ck(f(['uint144'], D), Ti)  # var case 76
ck(f(['uint152'], D), Ti)  # var case 77
ck(f(['uint16'], D), Ti)  # var case 78
ck(f(['uint160'], D), Ti)  # var case 79
ck(f(['uint168'], D), Ti)  # var case 80
ck(f(['uint176'], D), Ti)  # var case 81
ck(f(['uint184'], D), Ti)  # var case 82
ck(f(['uint192'], D), Ti)  # var case 83
ck(f(['uint200'], D), Ti)  # var case 84
ck(f(['uint208'], D), Ti)  # var case 85
ck(f(['uint216'], D), Ti)  # var case 86
ck(f(['uint224'], D), Ti)  # var case 87
ck(f(['uint232'], D), Ti)  # var case 88
ck(f(['uint24'], D), Ti)  # var case 89
ck(f(['uint240'], D), Ti)  # var case 90
ck(f(['uint248'], D), Ti)  # var case 91
ck(f(['uint256'], D), Ti)  # var case 92
ck(f(['uint32'], D), Ti)  # var case 93
ck(f(['uint40'], D), Ti)  # var case 94
ck(f(['uint48'], D), Ti)  # var case 95
ck(f(['uint56'], D), Ti)  # var case 96
ck(f(['uint64'], D), Ti)  # var case 97
ck(f(['uint72'], D), Ti)  # var case 98
ck(f(['uint8'], D), Ti)  # var case 99
ck(f(['uint80'], D), Ti)  # var case 100
ck(f(['uint88'], D), Ti)  # var case 101
ck(f(['uint96'], D), Ti)  # var case 102
ck(f(['?'], D), TA)  # var case 103



from importlib.metadata import (
    version as __version,
)

from faster_eth_abi.abi import (
    decode,
    encode,
    is_encodable,
    is_encodable_type,
)

__version__ = __version("faster-eth-abi")



from _typeshed import Incomplete
from faster_eth_abi.abi import (
    decode as decode,
    encode as encode,
    is_encodable as is_encodable,
    is_encodable_type as is_encodable_type,
)

__version__: Incomplete



import pytest
import decimal
import itertools
from typing import (
    Iterable,
    Iterator,
    TypeVar,
)

from hypothesis import (
    given,
    strategies as st,
)

from faster_eth_abi.utils.numeric import (
    ABI_DECIMAL_PREC,
    scale_places,
)

T = TypeVar("T")


def take(n: int, it: Iterable[T]) -> Iterator[T]:
    """
    Returns an iterator that takes ``n`` items from the iterable ``it``.
    """
    yield from (i for _, i in zip(range(n), it))


def get_test_value(prec: int, frac_places: int) -> decimal.Decimal:
    """
    Returns a test decimal value with ``prec`` total significant figures and
    ``frac_places`` fractional places.
    """
    digits = itertools.cycle(map(str, range(1, 10)))
    decimal_repr = "".join(take(prec, digits))

    if frac_places <= 0:
        # Add zeroes to right side of repr
        return decimal.Decimal(decimal_repr + "0" * -frac_places)
    elif frac_places > prec:
        # Add zeroes to left side of repr
        extra_frac_places = frac_places - prec
        return decimal.Decimal("." + "0" * extra_frac_places + decimal_repr)
    else:
        # Split repr according to frac places
        whole_places = prec - frac_places
        return decimal.Decimal(
            decimal_repr[:whole_places] + "." + decimal_repr[-frac_places:]
        )


@given(st.integers(min_value=1, max_value=ABI_DECIMAL_PREC))
def test_scale_places_pos(size):
    scaled_by_string = get_test_value(size, size)

    dec_value = get_test_value(size, 0)
    scaled_by_math = scale_places(size)(dec_value)

    assert scaled_by_string == scaled_by_math


@given(st.integers(min_value=1, max_value=ABI_DECIMAL_PREC))
def test_scale_places_neg(size):
    scaled_by_string = get_test_value(size, -size)

    dec_value = get_test_value(size, 0)
    scaled_by_math = scale_places(-size)(dec_value)

    assert scaled_by_string == scaled_by_math


def test_scale_places_raises_value_error_if_places_not_int():
    with pytest.raises(TypeError):
        scale_places(1.0)



import pytest

from faster_eth_abi.utils.string import (
    abbr,
)


@pytest.mark.parametrize(
    "value,expected,limit",
    (
        (1234567891234567891, "1234567891234567891", None),
        (12345678912345678912, "12345678912345678912", None),
        (
            "asdf" * 30,
            "'asdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasdfasd...",  # noqa: E501
            None,
        ),
        (
            list(range(100)),
            "[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 2...",  # noqa: E501
            None,
        ),
        (1234567891234567891, "...", 3),
        (1234567891234567891, "1...", 4),
    ),
)
def test_abbr(value, expected, limit):
    if limit is not None:
        actual = abbr(value, limit)
    else:
        actual = abbr(value)
    assert actual == expected


def test_abbr_throws_value_errors():
    with pytest.raises(ValueError):
        abbr("asdf", limit=2)



from ._grammar import (
    ABIType as ABIType,
    BasicType as BasicType,
    TupleType as TupleType,
    normalize as normalize,
)
from .base import BaseCoder as BaseCoder
from .grammar import parse as parse
from _typeshed import Incomplete
from typing import Callable, TypeVar

TType = TypeVar("TType", bound=type["BaseCoder"])
OldFromTypeStr: Incomplete
NewFromTypeStr: Incomplete

def parse_type_str(
    expected_base: str | None = None, with_arrlist: bool = False
) -> Callable[[OldFromTypeStr[TType]], "NewFromTypeStr[TType]"]:
    """
    Used by BaseCoder subclasses as a convenience for implementing the
    ``from_type_str`` method required by ``ABIRegistry``.  Useful if normalizing
    then parsing a type string with an (optional) expected base is required in
    that method.
    """

def parse_tuple_type_str(
    old_from_type_str: OldFromTypeStr[TType],
) -> NewFromTypeStr[TType]:
    """
    Used by BaseCoder subclasses as a convenience for implementing the
    ``from_type_str`` method required by ``ABIRegistry``.  Useful if normalizing
    then parsing a tuple type string is required in that method.
    """



"""
parse_benchmark_output.py

Extracts per-function benchmark timings from pytest-benchmark's benchmark.json output.
Parses the JSON file, finds all test function results, and writes a JSON file
mapping submodules (e.g., Python submodules like faster_eth_abi.encoding) to groups and their test functions and results.

Usage:
    python parse_benchmark_output.py <benchmark.json> [output.json]
"""

from collections import (
    defaultdict,
)
import json
import re
import sys
from typing import (
    Any,
    Dict,
)


def get_submodule(bench: dict) -> str:
    # Extract Python submodule from fullname, e.g., "benchmarks/test_abi_benchmarks.py::test_abi_to_signature"
    fullname = bench.get("fullname", "")
    # Try to extract the submodule from the test file path
    m = re.search(r"benchmarks/test_([a-zA-Z0-9_]+)_benchmarks\.py", fullname)
    if m:
        return f"faster_eth_abi.{m.group(1)}"
    return "unknown"


def get_group_name(test_name: str) -> str:
    # Extract group from test name, e.g., test_foo, test_faster_foo -> group: foo
    m = re.match(r"test_faster_(.+)", test_name)
    if m:
        return m.group(1)
    m = re.match(r"test_(.+)", test_name)
    if m:
        return m.group(1)
    return test_name


def parse_pytest_benchmark_json(data: dict) -> Dict[str, Dict[str, Dict[str, Any]]]:
    """
    Parses pytest-benchmark's benchmark.json and extracts per-function timings,
    grouped by submodule and group name.
    Returns a dict: {submodule: {group: {function_name: {...}}}}
    """
    results = defaultdict(lambda: defaultdict(dict))
    for bench in data.get("benchmarks", []):
        name = bench["name"]
        submodule = get_submodule(bench)
        group = get_group_name(name)
        stats = bench["stats"]
        results[submodule][group][name] = {
            "mean": stats.get("mean"),
            "stddev": stats.get("stddev", None),
            "iqr": stats.get("iqr", None),
            "min": stats.get("min", None),
            "max": stats.get("max", None),
            "rounds": stats.get("rounds", None),
        }
    return results


def main() -> None:
    if len(sys.argv) < 2:
        print("Usage: python parse_benchmark_output.py <benchmark.json> [output.json]")
        sys.exit(1)
    infile = sys.argv[1]
    outfile = sys.argv[2] if len(sys.argv) > 2 else "benchmark_results.json"
    with open(infile) as f:
        data = json.load(f)
    results = parse_pytest_benchmark_json(data)
    with open(outfile, "w") as f:
        json.dump(results, f, indent=2)
    print(f"Parsed results written to {outfile}")


if __name__ == "__main__":
    main()
